"""Embedding модели Compressa"""

from typing import List, Any, Dict, Optional
import os

from langchain_core.embeddings import Embeddings
from langchain_core.pydantic_v1 import Field, SecretStr, BaseModel
from langchain_openai import OpenAIEmbeddings

COMPRESSA_API_BASE = "https://compressa-api.mil-team.ru/v1"


class CompressaEmbeddings(BaseModel, Embeddings):
    """CompressaEmbeddings модели embedding.

    Для использования вам необходимо иметь переменную окружения ``COMPRESSA_API_KEY`` с установленным вашим ключом Compressa API, 
    или передать ключ как именованный параметр конструктора.

    Пример:
        .. code-block:: python

            from langchain_compressa import CompressaEmbeddings

            model = CompressaEmbeddings()
    """
    model: str = "/app/resources/models/models/Salesforce_SFR-Embedding-Mistral"
    tiktoken_enabled: bool = False
    tiktoken_model_name: Optional[str] = "Salesforce/SFR-Embedding-Mistral"
    model_kwargs: Dict[str, Any] = Field(default={"encoding_format": "float"})
    compressa_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    client: Any = Field(default=None, exclude=True)
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        compressa_api_key = self.compressa_api_key or SecretStr(os.getenv("COMPRESSA_API_KEY"))
        
        if compressa_api_key is None:
            raise Exception("status_code: None, body: The client must be instantiated be either passing in api_key or setting COMPRESSA_API_KEY")
            
        self.client = OpenAIEmbeddings(
            model=self.model,
            openai_api_base=COMPRESSA_API_BASE,
            openai_api_key=compressa_api_key,
            model_kwargs=self.model_kwargs,
            tiktoken_enabled=self.tiktoken_enabled,
            tiktoken_model_name=self.tiktoken_model_name,
        )

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Embed для списка документов"""

        return self.client.embed_documents(texts)

    def embed_query(self, text: str) -> List[float]:
        """Embed для строкового запроса"""
        
        return self.client.embed_query(text)



"""Rerank модели Compressa"""

from __future__ import annotations

import os
import requests
from copy import deepcopy
from typing import Any, Dict, List, Optional, Sequence, Union

from langchain_core.callbacks.manager import Callbacks
from langchain_core.documents import BaseDocumentCompressor, Document
from langchain_core.pydantic_v1 import Field, SecretStr


_RerankRequestDocumentsItem = Union[str, Dict]

class _CompressaClient:
    """
    Параметры:
    ----------
    base_url : typing.Optional[str]
        Базовый URL-адрес, который будет использоваться для запросов от клиента.

    compressa_api_key : typing.Optional[str]
        Ключ Compressa API.
    """

    def __init__(
        self,
        *,
        base_url: Optional[str] = os.getenv("COMPRESSA_BASE_URL", "https://compressa-api.mil-team.ru/v1"),
        compressa_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    ):
        
        self.compressa_api_key = compressa_api_key
        self.base_url = base_url

    def _rerank(
        self,
        *,
        query: str,
        documents: Sequence[_RerankRequestDocumentsItem],
        model: Optional[str] = "mixedbread-ai/mxbai-rerank-large-v1",
        top_n: Optional[int] = 5,
        return_documents: Optional[bool] = False
    ) -> any:  #TODO: обработать ответ RerankResponse 
        
        headers = {
            "Content-Type": "application/json",
            "X-Fern-Language": "Python",
            "Authorization": "Bearer " + self.compressa_api_key.get_secret_value()
        }
           
        jsonBody = {
            "model": model,
            "query": query,
            "documents": documents,
            "top_n": top_n,
            "return_documents": return_documents,
            }
	    
        _response = requests.post(f"{self.base_url}/rerank", headers=headers, json=jsonBody)

        if _response.status_code == 200:
            return _response.json() 
        else:
            raise Exception(f"status_code: {_response.status_code}, body: {_response.text}")


class CompressaRerank(BaseDocumentCompressor):
    """Компрессор документов который использует `Compressa Rerank API`."""

    top_n: Optional[int] = 3
    """Количество возвращаемых документов."""
    model: str = "mixedbread-ai/mxbai-rerank-large-v1"
    """Модель Compressa, используемая для реранка"""
    compressa_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    """Ключ Compressa API. Может быть определён непосредственно или путём установки переменной окружения COMPRESSA_API_KEY."""
    client: Any = Field(default=None, exclude=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        compressa_api_key = self.compressa_api_key or SecretStr(os.getenv("COMPRESSA_API_KEY"))
        
        if compressa_api_key is None:
            raise Exception("status_code: None, body: The client must be instantiated be either passing in api_key or setting COMPRESSA_API_KEY")
            
        self.client = _CompressaClient(compressa_api_key=compressa_api_key)

    def _rerank(
        self,
        documents: Sequence[Union[str, Document, dict]],
        query: str,
        *,
        model: Optional[str] = None,
        top_n: Optional[int] = -1,
    ) -> List[Dict[str, Any]]:
        """Возвращает список документов, упорядоченный по их релевантности предоставленному запросу.

        Args:
            query: Запрос, используемый для реранка.
            documents: Последовательность документов для реранка.
            model: Модель, используемая для реранкинга. По умолчанию self.model.
            top_n : Количество результатов для возвращения. Если None возвращаются все результаты.
                По умолчанию self.top_n.
        """
        if len(documents) == 0:  # чтобы избегать пустых вызовов API
            return []
        docs = [
            doc.page_content if isinstance(doc, Document) else doc for doc in documents
        ]
        model = model or self.model
        
        top_n = top_n if (top_n is None or top_n > 0) else self.top_n
        results = self.client._rerank(
            query=query,
            documents=docs,
            model=model,
            top_n=top_n,
        )
        result_dicts = []
        for res in results["results"]:
            result_dicts.append(
                {"index": res["index"], "relevance_score": res["relevance_score"]}
            )
        return result_dicts

    def compress_documents(
        self,
        documents: Sequence[Document],
        query: str,
        callbacks: Optional[Callbacks] = None,
    ) -> Sequence[Document]:
        """
        Сжатие (compress) документов с использованием Compressa rerank API.

        Args:
            documents: Последовательность документов для сжатия.
            query: Запрос, используемый для компрессии документов.
            callbacks: Обратные вызовы для запуска во время процесса сжатия.

        Returns:
            Последовательность сжатых документов.
        """
        compressed = []
        for res in self._rerank(documents, query):
            doc = documents[res["index"]]
            doc_copy = Document(doc.page_content, metadata=deepcopy(doc.metadata))
            doc_copy.metadata["relevance_score"] = res["relevance_score"]
            compressed.append(doc_copy)
        return compressed



"""Чат модели Compressa"""

from typing import Any, List, Optional, Dict, Iterator
import os

from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import BaseMessage
from langchain_core.outputs import ChatGenerationChunk, ChatResult
from langchain_core.pydantic_v1 import Field, SecretStr
from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_openai import ChatOpenAI

COMPRESSA_API_BASE = "https://compressa-api.mil-team.ru/v1"

class ChatCompressa(BaseChatModel):
    """Интеграция чат моделей.

    Установка:
        Установите пакет``langchain_compressa`` и переменную окружения ``COMPRESSA_API_KEY``.

        .. code-block:: bash

            pip install git+https://github.com/insight-stream/langchain_compressa.git
            export COMPRESSA_API_KEY="ваш-ключ-здесь"

    Ключевые аргументы инициализации — параметры для комплишн:
        model: str
            Имя модели Compressa для использования.
        temperature: float
            Температура выборки.

    Ключевые аргументы инициализации — параметры клиента
        api_key: Optional[str]
            Ключ Compressa API. Если не передан то будет читаться из переменной окружения COMPRESSA_API_KEY.

    Полный список поддерживаемых аргументов инициализации и их описания см. в разделе параметров.

    Создание экземпляра:
        .. code-block:: python

            from langchain_compressa import ChatCompressa

            llm = ChatCompressa(
                model="...",
                temperature=0,
                # api_key="...",
            )

    Вызов:
        .. code-block:: python

            messages = [
                ("system", "Ты полезный переводчик. Переведи предложение пользователя на французский."),
                ("human", "Я люблю программирование."),
            ]
            llm.invoke(messages)

    Стриминг:
        .. code-block:: python

            for chunk in llm.stream(messages):
                print(chunk)

            stream = llm.stream(messages)
            full = next(stream)
            for chunk in stream:
                full += chunk
            full
   
    """ 
    
    model_name: str = Field(default="/app/resources/models/models/compressa-ai_Llama-3-8B-Instruct", alias="model")
    """Имя чат модели для использования."""
    temperature: float = 0.7
    """Температура для использования."""
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    """Содержит любые параметры модели, действительные для вызова create, которые не указаны явно."""
    compressa_api_key: Optional[SecretStr] = Field(default=None, alias="api_key")
    """Автоматически берётся из переменной окружения `COMPRESSA_API_KEY` если не предоставлен."""
    streaming: bool = False
    """Выполнять стриминг результатов или нет"""
    client: Any = Field(default=None, exclude=True)
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        
        compressa_api_key = self.compressa_api_key or SecretStr(os.getenv("COMPRESSA_API_KEY"))
        
        if compressa_api_key is None:
            raise Exception("status_code: None, body: The client must be instantiated be either passing in api_key or setting COMPRESSA_API_KEY")
            
        self.client = ChatOpenAI(
            model=self.model_name,
            temperature=self.temperature,
            base_url=COMPRESSA_API_BASE,
            api_key=compressa_api_key
        )

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        
        return self.client._generate(messages, stop, run_manager, **kwargs)
        

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
    
        return self.client._stream(messages, stop, run_manager, **kwargs)


    @property
    def _llm_type(self) -> str:
        """Возвращает тип чат модели."""
        return "chat-compressa"
        
    @property
    def _default_params(self) -> Dict[str, Any]:
        """Получить параметры по умолчанию для вызова Compressa API."""
        params = {
            "model": self.model_name,
            "stream": self.streaming,
            "temperature": self.temperature,
            **self.model_kwargs,
        }
        return params
        
    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Получить идентифицирующие параметры."""
        return {"model_name": self.model_name, **self._default_params}



from importlib import metadata

from langchain_compressa.embeddings import CompressaEmbeddings
from langchain_compressa.chat_models import ChatCompressa
from langchain_compressa.reranks import CompressaRerank

try:
    __version__ = metadata.version(__package__)
except metadata.PackageNotFoundError:
    # Случай, когда метаданные пакета недоступны.
    __version__ = ""
del metadata  # необязательно, позволяет избежать загрязнения результатов dir(__package__)

__all__ = [
    "CompressaEmbeddings",
    "ChatCompressa",
    "CompressaRerank",
    "__version__",
]
